@@ -1,7 +1,6 @@
-import gi
 import os
 import sys
-import importlib.util
+import typing
 from importlib.machinery import PathFinder
 from pathlib import Path
 
@@ -12,11 +11,18 @@ del sys.modules["gi"]
 import gi
 
 
+if typing.TYPE_CHECKING:
+    # Stubs for type checking our overrides
+    def require_version(namespace: str, version: str) -> None:
+        pass
+
+
 class GstOverrideImport:
+
     def find_spec(self, fullname, path, target=None):
         if not fullname.startswith("gi.overrides"):
             return None
-        finder = importlib.machinery.PathFinder()
+        finder = PathFinder()
         # From find_spec the docs:
         # If name is for a submodule (contains a dot), the parent module is automatically imported.
         spec = finder.find_spec(
